import std.string;
import std.conv;
import std.asserterror;

struct SaoriRequest {
    char[][] arguments;
    char[][char[]] headers;
}
struct SaoriResponse {
    char[] result;
    char[][] values;
    char[][char[]] headers;
}
extern(C) {
    bool saori_load(char[] path);
    bool saori_unload();
    int saori_execute(SaoriRequest, inout SaoriResponse);
    int saori_getversion(SaoriRequest, inout SaoriResponse);
}

bool startWith(char[] s, char[] sub)
{
    if(s == null || sub == null) return false;
    if(s.length < sub.length) return false;
    return s[0..sub.length] == sub;
}

unittest {
    assert(startWith("unko", "un") == true);
    assert(startWith("unko", "unko") == true);
    assert(startWith("un", "unko") == false);
    assert(startWith("unko", null) == false);
    assert(startWith(null, "unko") == false);
    assert(startWith(null, null) == false);
}

char[] saori_request(char[] s)
in {
    assert(s != null);
}
out(result) {
    assert(result != null);
    assert(result[0..6] == "SAORI/");
}
body {
    SaoriRequest req; // = new SaoriRequest();
    SaoriResponse res;
    if(!parserequest(s, req)) {
        return buildResponse(400, res);
    }
    res.headers["Charset"] = req.headers["Charset"];
    int statusCode;
    try {
        if(startWith(s, "EXECUTE")) {
            statusCode = saori_execute(req, res);
        }
        else if(startWith(s, "GET Version")) {
            statusCode = saori_getversion(req, res);
        }
        else {
            statusCode = 400;
        }
    }
    catch(AssertError e) {
        // ܂O֌W Java ɋy΂ȂˁB
        e.print();
        statusCode = 500;
    }
    return buildResponse(statusCode, res);
}

/**
 * @brief 胊NGXg؂荏ށB
 * @param s NGXg
 * @param req ؂荏܂ꂽNGXg
 * @return true ł
 *         false tH[}bg
 */
private bool parserequest(char[] s, out SaoriRequest req)
in {
    assert(s != null);
}
body {
    char[][] lines = splitlines(s);
    const char[] delim = ": ";
    const char[] ARGUMENT = "Argument";
    foreach(char[] line; lines[1..lines.length]) {
        // sŏIB
        if(line.length == 0) break;
        // RTBΎB
        int colon = line.find(delim);
        if(colon < 0) {
            return false;
        }
        char[] name = line[0..colon];
        char[] value = line[colon+delim.length..line.length];
        req.headers[name] = value;
        if(startWith(name, ARGUMENT)) {
            try {
                uint index = toUint(name[ARGUMENT.length..name.length]);
                if(req.arguments.length <= index) {
                    req.arguments.length = index + 1;
                }
                req.arguments[index] = value;
            }
            catch(ConvError) {
                // ȂɂȂ
            }
        }
    }
    return true;
}

private char[][int] statusName;
/**
 * @brief 背X|X𕶎ɂB
 */
private char[] buildResponse(int statusCode, SaoriResponse res)
in {
}
out(result) {
    assert(result != null);
    // ȂƂ "SAORI/1.0\r\n\r\n" ȏ͂͂̕
    assert(result.length >= "SAORI/1.0\r\n\r\n".length);
}
body {
    if(statusName.length == 0) {
        statusName[200] = "OK";
        statusName[204] = "No Content";
        statusName[400] = "Bad Request";
        statusName[500] = "Internal Server Error";
    }
    char[] s;
    s ~= "SAORI/1.0 ";
    char[] statusString = statusName[statusCode];
    if(!statusString) {
        statusCode = 500;
        statusString = statusName[statusCode];
    }
    s ~= toString(statusCode);
    s ~= " ";
    s ~= statusString;
    s ~= "\r\n";
    s ~= "Result: " ~ res.result;
    s ~= "\r\n";
    foreach(int i, char[] v; res.values) {
        s ~= "Value" ~ toString(i);
        s ~= ": ";
        s ~= v;
        s ~= "\r\n";
    }
    foreach(char[] k; res.headers.keys) {
        s ~= k;
        s ~= ": ";
        s ~= res.headers[k];
        s ~= "\r\n";
    }
    s ~= "\r\n";
    return s;
}

